#!/usr/bin/env python
# -*- coding:utf-8 -*-

import pathlib as pt
import inspect
import pprint

import yaml
import cerberus
import platform

from ioproc.schemas import general_schema, action_schema, checkpoint_schema


__author__ = ["Benjamin Fuchs", "Judith Vesper"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Felix Nitsch", "Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class ConfigurationError(Exception):
    """
    Error raised by the configuration process.
    """
    pass


class ConfigDict(dict):

    def __getitem__(self, fieldname):
        """
        Access the config dictionary and retrieve data set by field name.
        If an KeyError is raised and the actionName is None, a ConfigurationError
        is raised, warning the user of an unsuccessful configuration.
        :param field name
        :return data from config dictionary which is accessed using the field name
        """
        try:
            return super().__getitem__(fieldname)
        except KeyError as e:
            s = inspect.stack()

            actionName = None
            lineno = s[0].lineno
            filename = pt.Path(s[0].filename).name

            for idx, iframe in enumerate(s):
                if iframe.function == '__actionwrapper__':
                    last = s[idx - 1]
                    actionName = last.function
                    lineno = last.lineno
                    filename = pt.Path(last.filename).name
                    break

            if actionName is None:
                out = '\n      config field "{}" unavailable\n'\
                      '      in file "{}" in line {}'.format(fieldname, filename, lineno)
            else:
                out = '\n      config field "{}" unavailable\n'\
                      '      requested by action "{}" in line {}\n'\
                      '      in file "{}"'.format(fieldname, actionName, lineno, filename)
            raise ConfigurationError(out)

    def print(self):
        pprint.pprint(self)


class ConfigList(list):
    def __getitem__(self, fieldname):
        """
        Access the config list and retrieve data set by field name.
        If an KeyError is raised and the actionName is None, a ConfigurationError
        is raised, warning the user of an unsuccessful configuration.
        :param field name
        :return data from config list which is accessed using the field name
        """
        try:
            return super().__getitem__(fieldname)
        except (TypeError, IndexError):
            s = inspect.stack()

            actionName = None
            lineno = s[0].lineno
            filename = pt.Path(s[0].filename).name

            for idx, iframe in enumerate(s):
                if iframe.function == '__actionwrapper__':
                    last = s[idx - 1]
                    actionName = last.function
                    lineno = last.lineno
                    filename = pt.Path(last.filename).name
                    break

            if actionName is None:
                out = '\n      element at position {} unavailable\n'\
                      '      in file "{}" in line {}'.format(fieldname, filename, lineno)
            else:
                out = '\n      element at position {} unavailable\n'\
                      '      requested by action "{}" in line {}\n'\
                      '      in file "{}"'.format(fieldname, actionName, lineno, filename)

            raise ConfigurationError(out)


def convertList(d):
    """
    When loading the .yaml, the user defined workflow is stored in a dictionary.
    This method converts the dictionary to a list.
    :return list with actions to execute in workflow
    """
    if not isinstance(d, ConfigDict):
        for ikey, ivalue in d.items():
            if hasattr(ivalue, 'keys'):
                d[ikey] = convertList(ivalue)
            elif not hasattr(ivalue, 'strip') and hasattr(ivalue, '__iter__'):
                d[ikey] = ConfigList(ivalue)
    return d


def loadAndValidate(confPath):
    """
    Loads the config file from a path provided and validates it against a provided schema. If the config file is
    empty, an empty dictionary is returned in order to comply with the following interfaces.
    :param confPath: path to config, schema: validation schema which ensure certain standards of configuration
    :return config dictionary which is validated against a provided schema
    """
    conf = yaml.load(open(confPath), Loader=yaml.Loader)

    if conf is None:
        return {}

    dirs = {}
    if 'directives' in conf:
        dirs = conf['directives']
        del conf['directives']

    validate_config(conf, confPath)

    conf = convertList(conf)

    for idir in dirs:
        for itag, ival in idir.items():
            if itag == 'config':
                isubConfigTag, isubConfigPath = ival
                if isubConfigTag in conf:
                    raise KeyError(f'sub config with name "{isubConfigTag}" already exists. Rename subconfig')
                conf[isubConfigTag] = yaml.load(open(isubConfigPath), Loader=yaml.Loader)

    return conf


def validate_config(conf, confPath):
    """
    Raises error when user config violates specified schema. First a 'general_schema' is checked, which assures
    that user.yaml header is set correctly and each action in the workflow is defined with a 'project' and 'call'.
    Then, specified validations for 'call' == 'checkpoint' and all other actions are performed. If any of the
    validations is not successful, the procedure crashes logging the Exception
    """

    v_general = cerberus.Validator(general_schema)
    v_action = cerberus.Validator(action_schema)
    v_checkpoint = cerberus.Validator(checkpoint_schema)

    checks = list()
    validation_violation = list()

    if v_general.validate(conf):
        checks.append(True)

        for item in conf['workflow']:
            wrapped_item = dict()
            wrapped_item['action'] = item

            if list(item.values())[0]['call'] == 'checkpoint':
                if v_checkpoint.validate(wrapped_item):
                    checks.append(True)
                else:
                    checks.append(False)
                    validation_violation.append(v_checkpoint.errors)
            else:
                if v_action.validate(wrapped_item):
                    checks.append(True)
                else:
                    checks.append(False)
                    validation_violation.append(v_action.errors)
    else:
        checks.append(False)
        validation_violation.append(v_general.errors)

    if len(validation_violation) > 0:
        raise ConfigurationError('in config file "{}":\n{}'.format(confPath, validation_violation))


class ConfigProvider:
    """
    The ConfigProvider triggers the reading and validation of the config file.
    For this purpose it sets the pathes, parses the config file for a user schema
    and triggers the validation process.
    """
    def __init__(self):
        self.config = None
        self.userConfigPath = None

    def setPathes(self, userConfigPath):
        self.userConfigPath = userConfigPath

    def parse(self):
        self.config = ConfigDict()
        self.config["user"] = loadAndValidate(self.userConfigPath)

        return self.config

    def get(self):
        if self.config is None:
            self.parse()
        return self.config


configProvider = ConfigProvider()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import inspect
import json
import pathlib as pt

import pandas as pd

from ioproc.logger import datalogger as dlog


__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek", "Jan Buschmann"]

__license__ = "MIT"
__version__ = "1.0.1"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class DataFieldError(Exception):
    """
    Error message raised from DataManager if the requested data set is not available.
    """
    pass


class DataDict(dict):
    """
    The DataDict is used to serialize data items for a key provided
    """
    def __setitem__(self, key, value):
        if not isinstance(value, pd.DataFrame) and not isinstance(value, pd.Series):
            raise DataFieldError('The value "{1}" of the dict entry "{0}" has to be a Dataframe or a Series.'.format(key, value))
        super().__setitem__(key, value)

    def dataSerialize(self, key):
        for ikey, ival in self.items():
            yield '{}/{}'.format(key, ikey), ival


class DataManager(dict):
    """
    The DataManager behaves like a dictionary.
    It stores input and output data during run time.
    It makes the data available by their keys and
    can be called at any time in the workflow.
    """
    _overwriteFlag = False

    class __OverwriteProxy:
        """
        Changes the overwriteFlag in the data manager temporarily in order to
        allow overwriting of already existing data sets using 'with' statements.
        """
        def __init__(self, mother):
            self.mother = mother

        def __enter__(self):
            self.mother._overwriteFlag = True
            dlog.debug('activated overwrite of data sets')

        def __exit__(self, exc_type, exc_val, exc_tb):
            dlog.debug('deactivated overwrite of data sets')
            self.mother._overwriteFlag = False

    def __init__(self, config, *args, **kwargs):
        """
        Initializes the DataManager with the default dict init.
        Then it sets the additional attribute _overwriteProxy
        """
        super().__init__(*args, **kwargs)
        self._overwriteProxy = self.__OverwriteProxy(self)
        self._actionCfg = None
        self._inAction = False
        self._accessLog = []
        self._config = config

    @property
    def overwrite(self):
        """
        Is used in a with block, to enable overwrites of data fields in the data manager.
        Otherwise an error is raised.
        :return: the __OverwriteProxy for overwrites
        """
        return self._overwriteProxy

    def entersAction(self, actionCfg):
        """
        Executed when an action will be executed.
        This process registers the action internally and sets the flag _inAction.
        :param actionCfg:
        """
        self._actionCfg = actionCfg
        self._inAction = True

    def leavesAction(self):
        """
        Executed when an action has been processed.
        This process signals the data manager that it can clear the access log and to
        reset the current action name _actionName as well as the flag _inAction.
        """
        self._actionCfg = None
        self._inAction = False
        self._accessLog.clear()

    def _addToAccessLog(self, dataSetName, modifier):
        """
        Adds a data set to the access log for reporting.
        :param dataSetName: the name of the data set accessed
        :param modifier: the modifier, so read or write access
        """
        assert modifier in ['r', 'w'], 'access log modifier needs to be read r or write w'
        self._accessLog.append(dataSetName+' ('+modifier+')')

    def report(self):
        """
        Creates a text report from the access log detailing which
        data sets were accessed and if for read or write purposes.
        :return:
        """
        ret = f'During execution of action "{self._actionCfg["call"]}" (project: "{self._actionCfg["project"]}") ' \
              f'the following data sets were accessed:\n'
        for i in self._accessLog:
            ret += '    '+i+'\n'
        return ret

    def __setitem__(self, fieldname, dataset):
        """
        Sets the data set specified by field name to value and
        checks if overwrite of existing data fields is enabled.
        Checks also if the field was specified as an output for the action.
        :param fieldname: the field name
        :param dataset: the data set
        """
        errorOccured = False
        if self._actionCfg['call'] not in ['checkpoint', ]:
            if self._actionCfg['data']['write_to_dmgr'] is None:
                errorOccured = True
            else:
                if fieldname not in self._actionCfg['data']['write_to_dmgr']:
                    errorOccured = True

        if errorOccured:
            print(self._config)
            if self._config['debug']['enable development mode']:
                dlog.warn(f'the dataset "{fieldname}" was not declared as output for action '
                      f'"{self._actionCfg["call"]}". Please check the workflow specified in the user.yaml!')
            else:
                raise IOError(f'the dataset "{fieldname}" was not declared as output for action '
                    f'"{self._actionCfg["call"]}". Please check the workflow specified in the user.yaml!')

        if self._overwriteFlag:
            dlog.info('overwriting data set "{}"'.format(fieldname))
        assert self._overwriteFlag or fieldname not in self.keys(), 'dataset "{}" already in data manager. ' \
                                                                    'Overwrite prohibited. Use with dmgr: block!'.format(fieldname)

        dlog.debug('setting field: ' + fieldname + '->' + str(dataset))
        self._addToAccessLog(fieldname, 'w')

        super().__setitem__(fieldname, dataset)

    def __getitem__(self, fieldname):
        """
        Retrieve data set by field name.
        Checks also if the field was specified as an input for the action.
        :param fieldname:
        """

        errorOccured = False
        if self._actionCfg['call'] not in ['checkpoint',]:
            if self._actionCfg['data']['read_from_dmgr'] is None:
                errorOccured = True
            else:
                if fieldname not in self._actionCfg['data']['read_from_dmgr']:
                    errorOccured = True

        if errorOccured:
            if self._config['debug']['enable development mode']:
                dlog.warn(f'the dataset "{fieldname}" was not declared as input for action '
                          f'"{self._actionCfg["call"]}". Please check the workflow specified in the user.yaml!')
            else:
                raise IOError(f'the dataset "{fieldname}" was not declared as input for action '
                              f'"{self._actionCfg["call"]}". Please check the workflow specified in the user.yaml!')


        self._addToAccessLog(fieldname, 'r')
        dlog.debug('getting field: ' + fieldname)
        try:
            return super().__getitem__(fieldname)
        except KeyError as e:
            s = inspect.stack()

            actionName = 'undefined'
            lineno = 'MIA'
            filename = 'MIA'
            for idx, iframe in enumerate(s):
                if iframe.function == '__actionwrapper__':
                    last = s[idx-1]
                    actionName = last.function
                    lineno = last.lineno
                    filename = pt.Path(last.filename).name
                    break

            raise DataFieldError('\n      dataset "{}" unavailable\n'
                                 '      requested by action "{}" in line {}\n'
                                 '      in file "{}"'.format(fieldname, actionName, lineno, filename))

    def validate(self, reference, raiseError=True):
        """
        Validate if fields listed in reference exist in the data manager
        :param reference: a list of data field names
        :return: True/False or raises DataFieldError
        """
        hasErrors = False
        errorCode = ''
        for field in reference:
            if field not in self:
                hasErrors = True
                errorCode += ' ' + field
        if hasErrors and raiseError:
            raise DataFieldError('Missing fields in DataManager: ' + errorCode)

        return hasErrors

    def toCache(self, tag):
        """
        Writes the current content of the DataManager (only DataFrames, Series and serializable .json data)
        in .hdf5 format to disc with filename Cache_[tag], so that the work can be resumed at a later point. Overwrites
        existing cache files sharing the same tag.
        :param tag: user specified filename extension
        """
        with pd.HDFStore('./Cache_{}.h5f'.format(tag), mode='w') as cache:
            otherData = pd.Series()
            for ikey, ivalue in self.items():
                if isinstance(ivalue, pd.DataFrame) or isinstance(ivalue, pd.Series):
                    cache.put(ikey, ivalue)
                elif hasattr(ivalue, 'items'):
                    ivalue = DataDict(ivalue)
                    for k, v in ivalue.dataSerialize(ikey):
                        cache.put(k, v)
                else:
                    try:
                        otherData.loc[ikey] = json.dumps(ivalue)
                    except:
                        otherData.loc[ikey] = 'missing'
                        dlog.warning(
                            'Dataset "{}" could not be saved to cache. Unhandled data type "{}"'.format(ikey, type(ivalue)))
                        dlog.exception()
            cache.put('__otherData', otherData)

    def fromCache(self, tag, iActionInfo):
        """
        Clears the current DataManager, reads the cached data from disc in .hdf5 format
        and loads it into the DataManager.
        :param tag: filename extension of the file to read
        :param iActionInfo: current action to execute
        """
        self.clear()
        self._actionCfg = iActionInfo

        with pd.HDFStore('./Cache_{}.h5f'.format(tag), mode='r+') as cache:
            if '__otherData' in cache.keys():
                for ikey, ivalue in cache['__otherData'].items():
                    self[ikey] = json.loads(ivalue)

            for ikey in cache:
                ikey = ikey.lstrip('/')
                if '/' in ikey:
                    key0, key1 = ikey.split('/')
                    if key0 not in self:
                        self[key0] = {}
                    self[key0][key1] = cache[ikey]
                    continue
                if ikey == '__otherData': continue
                self[ikey] = cache[ikey]



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import os
import pathlib as pt

from ioproc.driver import _main


__author__ = ["Benjamin Fuchs"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Felix Nitsch", "Judith Vesper", "Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


envPath = pt.Path()

pathvar = os.environ['PATH']
elem = pathvar.split(';')

for ielem in elem:
    if 'Scripts' in ielem:
        envPath = pt.Path(ielem).parent
        break


def create_folders(workflowName='yourTestWorkflow'):
    '''
    Creates the required folder structure.
    '''
    _main(setupworkflow=False, userconfig=None, setupfolders=True, workflowname=workflowName)

def create_workflow(workflowName='yourTestWorkflow'):
    '''
    Creates a new workflow in the current work directory.
    '''
    # REMINDER: currently not used. Could be interfaced in the future.
    _main(setupworkflow=True, userconfig=None, setupfolders=False, workflowname=workflowName)

def start(userconfig=None):
    '''
    Executes the workflow manager.
    '''
    _main(setupworkflow=False, userconfig=userconfig, setupfolders=False, workflowname=None)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from functools import wraps
import shutil
import pathlib as pt

import arrow as ar
from frozendict import frozendict
import yaml

from ioproc.defaults import *
from ioproc.actionmanager import getActionManager
from ioproc.datamanager import DataManager
from ioproc.logger import mainlogger as log


__author__ = ["Benjamin Fuchs", "Judith Vesper", "Kai von Krbek"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Felix Nitsch", "Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


def setupFolderStructure(workflowName=''):
    """
    Set up folder structure, which is required for use of the ioProc workflow manager.
    If path already exists, a warning is raised.
    """
    if workflowName.strip() == '':
        workflowName = 'workflow1'

    workflowName = workflowName.replace(' ', '_')

    CWD = pt.Path.cwd()

    path2actions = pt.Path(CWD, 'actions', 'general.py')
    path2workflow = pt.Path(CWD, 'workflows', workflowName)
    path2userConfig = pt.Path(path2workflow, 'user.yaml')

    path2actions.parent.mkdir(parents=True, exist_ok=True)
    path2workflow.mkdir(parents=True, exist_ok=True)

    if path2userConfig.exists():
        log.warning(f'UserConfig already exists at location {path2userConfig}')
    else:
        try:
            with path2userConfig.open('w') as opf:
                opf.write(defaultUserContent)

            with path2userConfig.open('r') as ipf:
                d = yaml.load(ipf, Loader=yaml.SafeLoader)

            with path2userConfig.open('w') as opf:
                d['actionFolder'] = path2actions.parent.as_posix()
                yaml.dump(d, opf)
        except Exception as e:
            shutil.rmtree(path2userConfig.as_posix())
            raise e

    if path2actions.exists():
        log.warning(f'General.py already exists at location {path2actions.parent}')
    else:
        with path2actions.open('w') as opf:
            opf.write(defaultActions)

    root = pt.Path().cwd()
    folderScript = root / 'create_folder_structure.py'
    if not folderScript.exists():
        with folderScript.open('w') as opf:
            opf.write(defaultSetupFoldersPyFile)

    workflowStartScript = path2workflow / 'run.py'
    if not workflowStartScript.exists():
        with workflowStartScript.open('w') as opf:
            opf.write(defaultRunPyFile)


class action:
    def __init__(self, project):
        """
        decorator to register actions in the ActionManager and validate input
        :param project: name of project
        """
        self.projectName = project

    def __call__(self, f):
        actionMgr = getActionManager()

        @wraps(f)
        def __actionwrapper__(dmgr, config, params):
            log.debug('calling wrapper')

            if not type(dmgr) == DataManager: raise TypeError('no valid dataManager instance passed')

            start = ar.now()

            try:
                ret = f(dmgr, config, params)
            except Exception as e:
                log.exception(f'Error occured calling action "{self.projectName}-{f.__name__}"')
                raise e

            end = ar.now()
            log.info(end-start)
            return ret

        actionMgr[self.projectName, f.__name__] = __actionwrapper__

        log.debug('wrapping complete for action: '+str(__actionwrapper__.__name__))
        return __actionwrapper__



def freeze(d):
    '''
    This function converts a given dictionary in a read-only mapping, also called frozen-dictionary.
    :return a frozen dictionary where a key cannot be added nor removed
    '''
    for key, value in d.items():
        if not hasattr(value, 'lstrip') and not hasattr(value, 'items') and hasattr(value, '__iter__'):
            d[key] = tuple(value)
        elif hasattr(value, 'items'):
            d[key] = freeze(value)
    return frozendict(d)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
import importlib.util
import pathlib as pt

from ioproc.config import configProvider


__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class ActionError(Exception):
    '''
    Error raised by the ActionManager
    '''
    pass


class ActionManager(dict):
    '''
    Behaves like a dictionary and stores the project name and the action name together with a reference to the action.
    '''

    def __setitem__(self, key, value):
        '''
        overloaded method to handle two level keys.
        :param key: a tuple containing the project name and the action name
        :param value: the reference to an action
        '''
        projectName, actionName = key

        if projectName not in self:
            super().__setitem__(projectName, {})
            
        if actionName in self[projectName]:
            raise ActionError("Action {} already registered.".format(actionName))

        self[projectName][actionName] = value


class __actioninit:
    '''
    Asserts that only one ActionManager instance exists in the framework (Singleton).
    Within the framework, getActionManager is called to get a new ActionManager instance.
    getActionManager is an instance of the __actioninit class. On calling the getActionManager
    a check is done, if an ActionManager was already created. In this case the ActionManager
    instance is returned. If there is no instantiated ActionManager, a new instance is created
    and populated with actions.
    '''

    actionMgr = None

    def __call__(self):
        '''
        A call of the class __actioninit checks if an action manager instance is already available. (callable class)
        If no, it instantiates and populates a new one.
        If yes, it returns the previously created instance.
        :return: populated ActionManager instance
        '''
        if self.actionMgr is None:
            self.actionMgr = ActionManager()
            config = configProvider.get()

            for ifile in pt.Path(config['user']['actionFolder']).glob('./*.py'):
                spec = importlib.util.spec_from_file_location(f"actions.{ifile.name.strip('.py')}", ifile.as_posix())
                __m = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(__m)

        return self.actionMgr

getActionManager = __actioninit()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__ = ["Benjamin Fuchs", "Jan Buschmann", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = []

__license__ = "MIT"
__version__ = "1.0.1"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


general_schema = {
    'workflow': {
        'required': True,
        'type': 'list',
        'schema': {
            'type': 'dict',
            'valueschema': {
                'type': 'dict',
                'schema': {
                    'project': {
                        'type': 'string',
                        'required': True,
                    },
                    'call': {
                        'type': 'string',
                        'required': True,
                    },
                    'data': {
                        'required': False,
                    },
                    'args': {
                        'required': False,
                    },
                    'tag': {
                        'required': False,
                    },
                },
           }
        }
    },
    'actionFolder': {
        'type': 'string',
        'required': True,
    },
    'debug': {
        'type': 'dict',
        'schema': {
            'timeit': {
                'type': 'boolean',
            },
            'enable development mode': {
                'type': 'boolean',
            }
        }
    },
    'fromCheckPoint': {
        'type': ['string', 'integer'],
    }
}


action_schema = {
    'action': {
        'type': 'dict',
        'keyschema': {
            'type': 'string',
        },
        'valueschema': {
            'type': 'dict',
            'schema': {
                'project': {
                    'type': 'string',
                    'required': True,
                },
                'call': {
                    'type': 'string',
                    'required': True,
                },
                'data': {
                    'required': True,
                    'schema': {
                        'read_from_dmgr': {
                            'nullable': True,
                            'type': ['string', 'list'],
                            'required': True,
                            'forbidden': ['None', 'none'],
                        },
                        'write_to_dmgr': {
                            'nullable': True,
                            'type': ['string', 'list'],
                            'required': True,
                            'forbidden': ['None', 'none'],
                        },
                    }
                },
                'args': {
                    'type': 'dict',
                    'required': False,
                }
            }
        }
    }
}


checkpoint_schema = {
    'action': {
        'type': 'dict',
        'keyschema': {
            'type': 'string',
        },
        'valueschema': {
            'type': 'dict',
            'schema': {
                'project': {
                    'type': 'string',
                    'required': True,
                },
                'call': {
                    'type': 'string',
                    'required': True,
                },
                'tag': {
                    'type': ['string', 'integer'],
                    'required': True,
                }
            }
        }
    }
}



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from setuptools import setup


__author__ = ["Benjamin Fuchs", "Judith Riehm", "Felix Nitsch", "Jan Buschmann"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "2.0.4"
__maintainer__ = "Felix Nitsch"
__email__ = "ioproc@dlr.de"
__status__ = "Production"


def readme():
      with open('README.md') as f:
            return f.read()


setup(name='ioproc',
      version='2.0.4',
      description='Framework for data pre- and postprocessing',
      long_description='''ioProc is a light-weight workflow manager for Python ensuring robust, scalable and 
                        reproducible data pipelines. The tool is developed at the German Aerospace Center (DLR) 
                        for and in the scientific context of energy systems analysis, however, it is widely 
                        applicable in other scientific fields.
                        ''',
      keywords='workflow management, data pipeline, data science',
      url='https://gitlab.com/dlr-ve/ioproc',
      author=', '.join(__author__),
      author_email=__email__,
      license=__license__,
      packages=['ioproc'],
      install_requires=['arrow', 'pandas', 'frozendict', 'pyyaml', 'click', 'cerberus', 'tables'],
      zip_safe=False,
      include_package_data=True,
      )



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import pathlib as pt

import click

from ioproc.config import configProvider
from ioproc.defaults import *
from ioproc.logger import mainlogger as log
from ioproc.datamanager import DataManager
from ioproc.actionmanager import getActionManager
from ioproc.tools import freeze, setupFolderStructure


__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.2"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


IOPROCINSTALLROOT = pt.Path(__file__).resolve().parent
SCHEMAPATH = pt.Path(IOPROCINSTALLROOT, "schema")
HOME = pt.Path.home()

defaultConfigContent = defaultConfigContent.format(IOPROCINSTALLROOT.as_posix())


@click.command()
@click.option('--setupworkflow', is_flag=True, help='generate default setupworkflow setup')
@click.option('--userconfig', '-c', default=None, help='path to user.yaml')
@click.option('--setupfolders', is_flag=True, help='generate default folder structure')
@click.option('--workflowname', default='workflow1', help='name of your workflow' )
def ioproc(setupworkflow, userconfig, setupfolders, workflowname):
    _main(setupworkflow, userconfig, setupfolders, workflowname)


def _main(setupworkflow=False, userconfig=None, setupfolders=False, workflowname='workflow1'):
    """
    Main driver which triggers the setup of the required folder structure, if needed.
    It also sets the user config path and the workflow structure, if they are not set up already.
    If the workflow is already set up, the action manager and the data manager are initialized.
    Next and according to the provided user configuration, the workflow with its actions
    is executed, which can also be started from a checkpoint to resume a previously started run.
    If loading from a checkpoint is enabled, it is ensured that the given tag is existent in the directory,
    otherwise an Exception is raised.
    """
    if setupfolders:
        setupFolderStructure(workflowname)
        return

    userConfigPath = pt.Path(pt.Path.cwd(), 'user.yaml')

    if userconfig is not None:
        userConfigPath = pt.Path(userconfig)

    if setupworkflow and not userConfigPath.exists():
        if not userConfigPath.parent.exists():
            raise IOError(f"Path to user config not found: {userConfigPath.as_posix()}")

        with userConfigPath.open('w') as opf:
            opf.write(defaultUserContent)

        workflowStartScript = userConfigPath / 'run.py'
        if not workflowStartScript.exists():
            with workflowStartScript.open('w') as opf:
                opf.write(defaultRunPyFile)

    configProvider.setPathes(
                             userConfigPath=userConfigPath,
                             )
    config = configProvider.get()

    if not setupworkflow:
        actionMgr = getActionManager()
        assert len(actionMgr) > 0, "ActionManager is not defined. Ensure 'actionFolder' path in 'user.yaml' is set correctly."
        dmgr = DataManager(config['user'])

        log.info('starting workflow')

        log.debug('commencing action calling')

        FROMCHECKPOINT = config['user']['fromCheckPoint'] != 'start'

        if FROMCHECKPOINT:
            given_tag = config['user']['fromCheckPoint']
            if not pt.Path('./Cache_{}.h5f'.format(given_tag)).is_file():
               message = 'cannot find cache file with given tag name `{}`'.format(given_tag)
               log.exception(message)
               raise Exception(message)

        for iActionInfo in config['user']['workflow']:
            iActionInfo = iActionInfo[list(iActionInfo.keys())[0]]
            if FROMCHECKPOINT and 'tag' in iActionInfo and iActionInfo['tag'] != config['user']['fromCheckPoint']:
                continue
            elif FROMCHECKPOINT and 'tag' in iActionInfo and iActionInfo['tag'] == config['user']['fromCheckPoint']:
                FROMCHECKPOINT = False
                dmgr.fromCache(config['user']['fromCheckPoint'], iActionInfo)
                log.info('reading from cache for tag "{}"'.format(config['user']['fromCheckPoint']))
                continue
            elif FROMCHECKPOINT:
                continue

            log.debug('executing action "'+iActionInfo['call']+'"')
            dmgr.entersAction(iActionInfo)
            try:
                actionMgr[iActionInfo['project']][iActionInfo['call']](dmgr, config, freeze(iActionInfo))
            except Exception as e:
                log.exception('Fatal error during execution of action "'+iActionInfo['call']+'":\nData manager log:\n'+dmgr.report())
                raise e
            dmgr.leavesAction()


if __name__ == '__main__':
    ioproc()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import ioproc.driver


__author__ = ["Benjamin Fuchs",]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Judith Vesper", "Felix Nitsch", "Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


ioproc.driver.ioproc()


#!/usr/bin/env python
# -*- coding:utf-8 -*-
import logging
import logging.handlers
import pathlib as pt
import warnings
import inspect


__author__ = ["Benjamin Fuchs",]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Felix Nitsch", "Judith Vesper", "Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class FancyLogger(logging.Logger):
    '''
    This is a special logger, that by default (aka after calling self.setupDefault(PATH_TO_LOG_FILE))
    will have a stream handler and rotating file handler both set to DEBUG and a special format
    containing both the filename and line number of the logging call.
    '''
        
    def info(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().info(m, **kwargs)

    def debug(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().debug(m, **kwargs)

    def error(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().error(m, **kwargs)

    def warn(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().warning(m, **kwargs)

    def warning(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().warning(m, **kwargs)

    def exception(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().exception(m, **kwargs)

    def log(self, level, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().log(level, m, **kwargs)

    def critical(self, *args, **kwargs):
        m = ['{}'] * len(args)
        m = (' '.join(m)).format(*args)
        super().critical(m, **kwargs)

    def _setLogRecordFactory(self):
        '''
        Required to inject the correct line and file from which
        the logging call was emitted.
        :return: a valid record factory
        '''
        old_factory = logging.getLogRecordFactory()
        def record_factory(*args, **kwargs):
            record = old_factory(*args, **kwargs)

            stack = inspect.stack()[5]

            record.filename = pt.Path(stack.filename).name
            record.lineno = stack.lineno
            return record
        logging.setLogRecordFactory(record_factory)

    def setupDefaultFormat(self):
        '''
        Setup of the default output format for the logger.
        '''
        self._setLogRecordFactory()
        fmt = logging.Formatter('[{levelname:_<8}][{asctime}][{filename:_>15}][l.{lineno:>3}][{name}]: {msg}', style='{')
        for ihandler in self.handlers:
            ihandler.setFormatter(fmt)

        if len(self.handlers) == 0:
            warnings.warn('no handler set for logger "{}". Call setupDefaultFormatter() _after_ adding handlers or call setupDefault()')

    def setupDefaultHandlers(self, fileHandlerPath, streamHandlerLevel=logging.DEBUG, fileHandlerLevel=logging.DEBUG):
        '''
        creates default file and stream handlers. File handler is a rotating one.
        :param fileHandlerPath: the path where files should be written to
        :param streamHandlerLevel: the log level for the console stream
        :param fileHandlerLevel: the log level for the log file
        '''
        sh = logging.StreamHandler()
        sh.setLevel(streamHandlerLevel)
        fh = logging.handlers.RotatingFileHandler(fileHandlerPath)
        fh.setLevel(fileHandlerLevel)
        self.addHandler(sh)
        self.addHandler(fh)
        self.setLevel(logging.DEBUG)

    def setupDefault(self, fileHandlerPath, streamHandlerLevel=logging.DEBUG, fileHandlerLevel=logging.DEBUG):
        '''
        convenience method to create a default logger fully configured.
        :param fileHandlerPath: the path where files should be written to
        :param streamHandlerLevel: the log level for the console stream
        :param fileHandlerLevel: the log level for the log file
        '''
        self.setupDefaultHandlers(fileHandlerPath, streamHandlerLevel, fileHandlerLevel)
        self.setupDefaultFormat()

logging.setLoggerClass(FancyLogger)

mainlogger = logging.getLogger('main')
mainlogger.setupDefault('./main.log')

datalogger = logging.getLogger('data')
datalogger.setupDefault('./datatrace.log')